from exceptionite.errors import Handler, StackOverflowIntegration, SolutionsIntegration

from .JsonHandler import JsonHandler


class ExceptionHandler:
    def __init__(self, application, driver_config=None):
        self.application = application
        self.drivers = {}
        self.driver_config = driver_config or {}
        self.options = {}

    def set_options(self, options):
        self.options = options
        return self

    def add_driver(self, name, driver):
        self.drivers.update({name: driver})

    def set_configuration(self, config):
        self.driver_config = config
        return self

    def get_driver(self, name=None):
        if name is None:
            return self.drivers[self.driver_config.get("default")]
        return self.drivers[name]

    def get_config_options(self, driver=None):
        if driver is None:
            return self.driver_config[self.driver_config.get("default")]

        return self.driver_config.get(driver, {})

    def handle(self, exception):
        response = self.application.make("response")
        request = self.application.make("request")

        self.application.make("event").fire(
            f"masonite.exception.{exception.__class__.__name__}", exception
        )

        if self.application.has(f"{exception.__class__.__name__}Handler"):
            return self.application.make(
                f"{exception.__class__.__name__}Handler"
            ).handle(exception)

        if hasattr(exception, "get_response"):
            return response.view(exception.get_response(), exception.get_status())

        handler = Handler(exception)
        if "application/json" in str(request.header("Accept")):
            return response.view(JsonHandler(exception).render(), status=500)

        if self.options.get("handlers.stack_overflow"):
            handler.integrate(StackOverflowIntegration())
        if self.options.get("handlers.solutions"):
            handler.integrate(SolutionsIntegration())

        handler.context(
            {
                "WSGI": {
                    "Path": request.get_path(),
                    "Input": request.input_bag.all_as_values() or None,
                    # 'Parameters': request.url_params,
                    "Request Method": request.get_request_method(),
                },
                "Headers": request.header_bag.to_dict(),
            }
        )

        return response.view(handler.render(), status=500)



from tests import TestCase
from src.masonite.request import Request
from src.masonite.utils.http import generate_wsgi
import os


class TestRequest(TestCase):
    def test_request_can_get_path(self):
        request = Request(generate_wsgi(path="/test"))
        self.assertEqual(request.get_path(), "/test")
        self.assertEqual(request.get_request_method(), "GET")

    def test_request_contains(self):
        request = Request(generate_wsgi(path="/test"))
        self.assertTrue(request.contains("/test"))

        request = Request(generate_wsgi(path="/test/user"))
        self.assertTrue(request.contains("/test/*"))

        request = Request(generate_wsgi(path="/test/admin/user"))
        self.assertTrue(request.contains("/test/*/user"))

        request = Request(generate_wsgi(path="/test/admin/user"))
        self.assertTrue(request.contains("*"))



from .exceptions import DumpException


class DD:
    def __init__(self, container):
        self.app = container

    def dump(self, *args):
        dump_list = []
        for i, obj in enumerate(args):
            dump_name = "ObjDump{}".format(i)
            self.app.bind(dump_name, obj)
            dump_list.append(dump_name)
        self.app.bind("ObjDumpList", dump_list)
        raise DumpException



import sys
import traceback
import pkg_resources
import math
import pprint


class JsonHandler:
    def __init__(self, e=False, **kwargs):
        self.e = e
        self._contexts = {}
        self._integrations = {}

        self.type, self.value, self.tb = (
            exc_type,
            exc_value,
            exc_traceback,
        ) = sys.exc_info()
        self.traceback = traceback.TracebackException(
            self.type, self.value, self.tb, capture_locals=True
        )
        self.trace = self.create_trace()

        self.python_version = (
            str(sys.version_info[0])
            + "."
            + str(sys.version_info[1])
            + "."
            + str(sys.version_info[2])
        )
        self.default_encoding = sys.getdefaultencoding()
        self.file_system_encoding = sys.getfilesystemencoding()
        self.platform = sys.platform

        self.context(
            {
                "Environment & Details": {
                    "Python Version": self.python_version,
                    "Platform": self.platform,
                    "File System Encoding": self.file_system_encoding,
                    "Default Encoding": self.default_encoding,
                    "Argv": sys.argv,
                }
            }
        )

        installed_packages = pkg_resources.working_set
        packages = {}
        for i in installed_packages:
            packages.update({i.key: i.version})

        self.context({"Packages": packages})

    def any(self):
        return bool(self.e)

    def count(self):
        return len(self.trace)

    def message(self):
        return str(self.e)

    def exception(self):
        return self.e.__class__.__name__

    def namespace(self):
        return self.e.__class__.__module__ + "." + self.exception()

    def stacktrace(self):
        traceback = self.trace
        traceback.reverse()
        return traceback

    def context(self, context: dict):
        self._contexts.update(context)
        return self

    def integrate(self, integration):
        if isinstance(integration, dict):
            self._integrations.update(integration)
        else:
            self.integrate(
                {
                    integration.name: {
                        "content": integration.content(self),
                        "cls": integration,
                    }
                }
            )
        return self

    def get_contexts(self):
        return self._contexts

    def get_integrations(self):
        return self._integrations

    def create_trace(self):
        traceback = []
        for tb in self.traceback.stack:
            traceback.append(SimpleStackLine(tb, variables=tb.locals))

        return traceback

    def render(self):
        data = {
            "exception": self.exception(),
            "message": self.message(),
            "file": self.stacktrace()[0].file,
            "trace": [],
        }

        for trace in self.trace:
            data["trace"].append(
                {
                    "file": trace.file,
                    "line": trace.lineno,
                    "statement": trace.parent_statement,
                }
            )

        return data


class SimpleStackLine:
    def __init__(self, frame_summary, variables={}):
        self.file = frame_summary[0]

        # Cut off 30% of the string
        self.file_short = ".." + self.file[math.floor(len(self.file) * 0.30) :]
        self.lineno = frame_summary[1]
        self.offending_line = self.lineno
        self.parent_statement = frame_summary[2]
        self.statement = frame_summary[3]
        self.start_line = self.lineno - 5
        self.end_line = self.lineno + 5
        self.variables = variables



"""Scaffold Auth Command."""
from distutils.dir_util import copy_tree
import os

from ..utils.location import controllers_path, views_path, mailables_path
from ..utils.filesystem import get_module_dir
from .Command import Command


class AuthCommand(Command):
    """
    Creates a new authentication scaffold.

    auth
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        copy_tree(
            self.get_template_path(),
            views_path("auth"),
        )
        copy_tree(self.get_controllers_path(), controllers_path("auth"))

        copy_tree(self.mailables_path(), mailables_path())

        self.info("Auth scaffolded successfully!")

    def get_template_path(self):
        return os.path.join(get_module_dir(__file__), "../stubs/templates/auth")

    def get_controllers_path(self):
        return os.path.join(get_module_dir(__file__), "../stubs/controllers/auth")

    def mailables_path(self):
        return os.path.join(get_module_dir(__file__), "../stubs/auth/mailables")



from tests import TestCase

from src.masonite.helpers import optional


class SomeClass:

    my_attr = 3

    def my_method(self):
        return 4


class TestOptionalHelper(TestCase):
    def test_optional_with_existing(self):
        obj = SomeClass()
        self.assertEqual(optional(obj).my_attr, 3)
        self.assertEqual(optional(obj).my_method(), 4)

    def test_optional_with_undefined(self):
        obj = SomeClass()
        self.assertEqual(optional(obj).non_existing_attr, None)
        self.assertEqual(optional(obj).non_existing_method(), None)

    def test_optional_with_default(self):
        obj = SomeClass()
        self.assertEqual(optional(obj, default=0).non_existing_attr, 0)
        self.assertEqual(optional(obj, default=0).non_existing_method(), 0)



from ..cookies import CookieJar
from ..headers import HeaderBag, Header
from ..input import InputBag
import re
import tldextract
from .validation import ValidatesRequest
from ..authorization import AuthorizesRequest
from ..sessions import old as old_helper


class Request(ValidatesRequest, AuthorizesRequest):
    def __init__(self, environ):
        """Request class constructor.

        Initializes several properties and sets various methods
        depending on the environtment.

        Keyword Arguments:
            environ {dictionary} -- WSGI environ dictionary. (default: {None})
        """
        self.environ = environ
        self.cookie_jar = CookieJar()
        self.header_bag = HeaderBag()
        self.input_bag = InputBag()
        self.params = {}
        self._user = None
        self._subdomains_activated = False
        self.load()

    def load(self):
        self.cookie_jar.load(self.environ.get("HTTP_COOKIE", ""))
        self.header_bag.load(self.environ)
        self.input_bag.load(self.environ)

    def load_params(self, params=None):
        if not params:
            params = {}

        self.params.update(params)

    def param(self, param, default=""):
        return self.params.get(param, default)

    def get_path(self):
        return self.environ.get("PATH_INFO")

    def get_path_with_query(self):
        return self.environ.get("PATH_INFO") + "?" + self.environ.get("QUERY_STRING")

    def get_back_path(self):
        return self.input("__back") or self.get_path_with_query()

    def get_request_method(self):
        return self.environ.get("REQUEST_METHOD")

    def input(self, name, default=""):
        """Get a specific input value.

        Arguments:
            name {string} -- Key of the input data

        Keyword Arguments:
            default {string} -- Default value if input does not exist (default: {False})
            clean {bool} -- Whether or not the return value should be
                            cleaned (default: {True})

        Returns:
            string
        """
        name = str(name)

        return self.input_bag.get(name, default=default)

    def cookie(self, name, value=None, **options):
        if value is None:
            cookie = self.cookie_jar.get(name)
            if not cookie:
                return
            return cookie.value

        return self.cookie_jar.add(name, value, **options)

    def delete_cookie(self, name):
        self.cookie_jar.delete(name)
        return self

    def header(self, name, value=None):
        if value is None:
            header = self.header_bag.get(name)
            if not header:
                return
            return header.value
        else:
            return self.header_bag.add(Header(name, value))

    def all(self):
        return self.input_bag.all_as_values()

    def only(self, *inputs):
        return self.input_bag.only(*inputs)

    def old(self, key):
        return old_helper(key)

    def is_not_safe(self):
        """Check if the current request is not a get request.

        Returns:
            bool
        """
        if not self.get_request_method() in ("GET", "OPTIONS", "HEAD"):
            return True

        return False

    def user(self):
        return self._user

    def set_user(self, user):
        self._user = user
        return self

    def remove_user(self):
        self._user = None
        return self

    def contains(self, route):
        if not route.startswith("/"):
            route = "/" + route

        regex = re.compile(route.replace("*", "[a-zA-Z0-9_]+"))

        return regex.match(self.get_path())

    def get_subdomain(self, exclude_www=True):
        if not self._subdomains_activated:
            return None

        url = tldextract.extract(self.get_host())
        if url.subdomain == "" or (
            url.subdomain and exclude_www and url.subdomain == "www"
        ):
            return None

        return url.subdomain

    def get_host(self):
        return self.environ.get("HTTP_HOST")

    def activate_subdomains(self):
        self._subdomains_activated = True
        return self



"""New Preset Command."""

from .Command import Command
from ..presets import Tailwind, Vue, React, Bootstrap, Remove


class PresetCommand(Command):
    """
    Scaffold frontend preset in your project

    preset
        {name? : Name of the preset [tailwind, vue, react, bootstrap]}
        {--r|remove=? : Remove all scaffolded presets}
    """

    presets = ["vue", "tailwind", "react", "bootstrap"]

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        preset_name = self.argument("name")
        remove_presets = self.option("remove")
        if remove_presets:
            return self.remove()

        if preset_name not in self.presets:
            return self.error(
                f"Invalid preset. Available presets are: {', '.join(self.presets)}"
            )

        self.info(f"Scaffolding {preset_name} preset...")
        return getattr(self, preset_name)()

    def remove(self):
        Remove().install()
        self.info("Frontend scaffolding removed successfully.")

    def react(self):
        React().install()
        self.info("React installed successfully.")
        self.comment(
            'Please run "npm install && npm run dev" to compile your fresh scaffolding.'
        )

    def vue(self):
        Vue().install()
        self.info("Vue 3.0 installed successfully.")
        self.comment(
            'Please run "npm install && npm run dev" to compile your fresh scaffolding.'
        )
        self.comment("Then you can use the view app_vue3 as demo.")

    def bootstrap(self):
        Bootstrap().install()
        self.info("Bootstrap installed successfully.")
        self.comment(
            'Please run "npm install && npm run dev" to compile your fresh scaffolding.'
        )

    def tailwind(self):
        Tailwind().install()
        self.info("TailwindCSS 3 installed successfully.")
        self.comment(
            'Please run "npm install && npm run dev" to compile your fresh scaffolding.'
        )



import json
import io
import sys
import unittest
import pendulum
from contextlib import contextmanager

from ..routes import Route
from ..foundation.response_handler import testcase_handler
from ..utils.http import generate_wsgi
from ..request import Request
from ..response import Response
from ..environment import LoadEnvironment
from ..providers.RouteProvider import RouteProvider
from .TestCommand import TestCommand


class TestCase(unittest.TestCase):
    routes_to_restore = ()

    def setUp(self):
        LoadEnvironment("testing")
        from wsgi import application

        self.application = application
        self.original_class_mocks = {}
        self._test_cookies = {}
        self._test_headers = {}
        if hasattr(self, "startTestRun"):
            self.startTestRun()
        self.withoutCsrf()
        self._exception_handling = False
        # boot providers as they won't not be loaded if the test is not doing a request
        self.application.bind("environ", generate_wsgi())
        try:
            for provider in self.application.get_providers():
                # if no request is made we don't need RouteProvider, and we can't load it anyway
                # because we don't have created a CSRF token yet
                if not isinstance(provider, RouteProvider):
                    application.resolve(provider.boot)
        except Exception as e:
            if not self._exception_handling:
                raise e
            self.application.make("exception_handler").handle(e)

        self.routes_to_restore = set(self.application.make("router").routes)

    def tearDown(self):
        # be sure to reset this between each test
        self._exception_handling = False
        # restore routes
        if self.routes_to_restore:
            self.application.make("router").routes = list(self.routes_to_restore)
        if hasattr(self, "stopTestRun"):
            self.stopTestRun()

    def withExceptionsHandling(self):
        """Enable for the duration of a test the handling of exceptions through the exception
        handler."""
        self._exception_handling = True

    def withoutExceptionsHandling(self):
        """Disable handling of exceptions."""
        self._exception_handling = False

    def setRoutes(self, *routes):
        """Set all routes of router during lifetime of a test."""
        self.application.make("router").set(Route.group(*routes, middleware=["web"]))
        return self

    def addRoutes(self, *routes):
        """Add routes to router during lifetime of a test."""
        self.application.make("router").add(Route.group(*routes, middleware=["web"]))
        return self

    def withCsrf(self):
        self._csrf = True
        return self

    def withoutCsrf(self):
        self._csrf = False
        return self

    def get(self, route, data=None):
        return self.fetch(route, data, method="GET")

    def post(self, route, data=None):
        return self.fetch(route, data, method="POST")

    def put(self, route, data=None):
        return self.fetch(route, data, method="PUT")

    def patch(self, route, data=None):
        return self.fetch(route, data, method="PATCH")

    def make_request(
        self, data={}, path="/", query_string="application=Masonite", method="GET"
    ):
        request = Request(generate_wsgi(data, path, query_string, method))
        request.app = self.application

        self.application.bind("request", request)
        return request

    def make_response(self, data={}):
        request = Response(generate_wsgi(data))
        request.app = self.application

        self.application.bind("response", request)
        return request

    def fetch(self, path, data=None, method=None):
        if data is None:
            data = {}
        if not self._csrf:
            token = self.application.make("sign").sign("cookie")
            data.update({"__token": "cookie"})
            wsgi_request = generate_wsgi(
                {
                    "HTTP_COOKIE": f"SESSID={token}; csrf_token={token}",
                    "CONTENT_LENGTH": len(str(json.dumps(data))),
                    "REQUEST_METHOD": method,
                    "PATH_INFO": path,
                    "wsgi.input": io.BytesIO(bytes(json.dumps(data), "utf-8")),
                }
            )
        else:
            wsgi_request = generate_wsgi(
                {
                    "CONTENT_LENGTH": len(str(json.dumps(data))),
                    "REQUEST_METHOD": method,
                    "PATH_INFO": path,
                    "wsgi.input": io.BytesIO(bytes(json.dumps(data), "utf-8")),
                }
            )

        request, response = testcase_handler(
            self.application,
            wsgi_request,
            self.mock_start_response,
            exception_handling=self._exception_handling,
        )
        # add eventual cookies added inside the test (not encrypted to be able to assert value ?)
        for name, value in self._test_cookies.items():
            request.cookie(name, value)
        # add eventual headers added inside the test
        for name, value in self._test_headers.items():
            request.header(name, value)

        route = self.application.make("router").find(path, method)
        if route:
            return self.application.make("tests.response").build(
                self.application, request, response, route
            )
        raise Exception(f"No route found for url {path}")

    def mock_start_response(self, *args, **kwargs):
        pass

    @contextmanager
    def captureOutput(self):
        new_out, new_err = io.StringIO(), io.StringIO()
        # save normal system output
        old_out, old_err = sys.stdout, sys.stderr
        try:
            sys.stdout, sys.stderr = new_out, new_err
            yield sys.stdout
        finally:
            # restore system output
            sys.stdout, sys.stderr = old_out, old_err

    def craft(self, command, arguments_str=""):
        """Run a given command in tests and obtain a TestCommand instance to assert command
        outputs.
        self.craft("controller", "Welcome").assertSuccess()
        """
        return TestCommand(self.application).run(command, arguments_str)

    def fake(self, binding):
        """Mock a service with its mocked implementation or with a given custom
        one."""

        # save original first
        self.original_class_mocks.update(
            {binding: self.application.make(binding, self.application)}
        )
        # mock by overriding with mocked version
        mock = self.application.make(f"mock.{binding}", self.application)
        self.application.bind(binding, mock)
        return mock

    def withCookies(self, cookies_dict):
        self._test_cookies = cookies_dict
        return self

    def withHeaders(self, headers_dict):
        self._test_headers = headers_dict
        return self

    def actingAs(self, user):
        self.make_request()
        self.application.make("auth").guard("web").login_by_id(
            user.get_primary_key_value()
        )

    def restore(self, binding):
        """Restore the service previously mocked to the original one."""
        original = self.original_class_mocks.get(binding)
        self.application.bind(binding, original)

    def fakeTime(self, pendulum_datetime):
        """Set a given pendulum instance to be returned when a "now" (or "today", "tomorrow",
        "yesterday") instance is created. It's really useful during tests to check
        timestamps logic."""
        pendulum.set_test_now(pendulum_datetime)

    def fakeTimeTomorrow(self):
        """Set the mocked time as tomorrow."""
        self.fakeTime(pendulum.tomorrow())

    def fakeTimeYesterday(self):
        """Set the mocked time as yesterday."""
        self.fakeTime(pendulum.yesterday())

    def fakeTimeInFuture(self, offset, unit="days"):
        """Set the mocked time as an offset of days in the future. Unit can be specified
        among pendulum units: seconds, minutes, hours, days, weeks, months, years."""
        self.restoreTime()
        datetime = pendulum.now().add(**{unit: offset})
        self.fakeTime(datetime)

    def fakeTimeInPast(self, offset, unit="days"):
        """Set the mocked time as an offset of days in the past. Unit can be specified
        among pendulum units: seconds, minutes, hours, days, weeks, months, years."""
        self.restoreTime()
        datetime = pendulum.now().subtract(**{unit: offset})
        self.fakeTime(datetime)

    def restoreTime(self):
        """Restore time to correct one, so that pendulum new "now" instance are corrects.
        This method will be typically called in tearDown() method of a test class."""
        # this will clear the mock
        pendulum.set_test_now()

    def assertDatabaseCount(self, table, count):
        self.assertEqual(self.application.make("builder").table(table).count(), count)

    def assertDatabaseHas(self, table, query_dict):
        self.assertGreaterEqual(
            self.application.make("builder").table(table).where(query_dict).count(), 1
        )

    def assertDatabaseMissing(self, table, query_dict):
        self.assertEqual(
            self.application.make("builder").table(table).where(query_dict).count(), 0
        )

    def assertDeleted(self, instance):
        self.assertFalse(
            self.application.make("builder")
            .table(instance.get_table_name())
            .where(instance.get_primary_key(), instance.get_primary_key_value())
            .get()
        )

    def assertSoftDeleted(self, instance):
        deleted_at_column = instance.get_deleted_at_column()
        self.assertTrue(
            self.application.make("builder")
            .table(instance.get_table_name())
            .where(instance.get_primary_key(), instance.get_primary_key_value())
            .where_not_null(deleted_at_column)
            .get()
        )



from .middleware_capsule import MiddlewareCapsule
from .middleware import Middleware
from .route.VerifyCsrfToken import VerifyCsrfToken
from .route.SessionMiddleware import SessionMiddleware
from .route.EncryptCookies import EncryptCookies
from .route.LoadUserMiddleware import LoadUserMiddleware
from .route.MaintenanceModeMiddleware import MaintenanceModeMiddleware
from .route.GuardMiddleware import GuardMiddleware



import inspect


class DumpExceptionHandler:
    def __init__(self, application):
        self.application = application

    def handle(self, exception):
        dump_objs = []
        for dump_name in self.application.make("ObjDumpList"):
            obj = self.application.make(dump_name)
            dump_objs.append(
                {
                    "obj": obj,
                    "members": inspect.getmembers(obj, predicate=inspect.ismethod),
                    "properties": inspect.getmembers(obj),
                }
            )

        return self.application.make("response").view(
            self.application.make("view").render(
                "/masonite/templates/dump",
                {
                    "objs": dump_objs,
                    "type": type,
                    "list": list,
                    "inspect": inspect,
                    "hasattr": hasattr,
                    "getattr": getattr,
                    "isinstance": isinstance,
                    "show_methods": (bool, str, list, dict),
                    "len": len,
                },
            )
        )



from os.path import join
from ..configuration import config


class UrlsHelper:
    """URLs helper provide handy functions to build URLs."""

    def __init__(self, app):
        self.app = app

    def url(self, path=""):
        """Generates a fully qualified url to the given path. If no path is given this will return
        the base url domain."""
        # ensure that no slash is prefixing the relative path
        relative_path = path.lstrip("/")
        return join(config("application.app_url"), relative_path)

    def asset(self, alias, filename):
        """Generates a fully qualified URL for the given asset using the given disk
        Example:
            asset("local", "avatar.jpg") (take first pat)
            asset("s3.private", "doc.pdf") (when multiple paths are specified for the disk)
        """
        disks = config("filesystem.disks")
        # ensure that no slash is prefixing the relative filename path
        filename = filename.lstrip("/")
        if "." in alias:
            alias = alias.split(".")
            location = disks[alias[0]]["path"][alias[1]]
        else:
            location = disks[alias]["path"]
            # take first path if no path specified
            if isinstance(location, dict):
                location = list(location.values())[0]
        return join(location, filename)

    def route(self, name, params={}, absolute=True):
        """Generates a fully qualified URL to the given route name.
        Example:
            route("users.home") : http://masonite.app/dashboard/
            route("users.profile", {"id": 1}) : http://masonite.app/users/1/profile/
            route("users.profile", {"id": 1}, absolute=False) : /users/1/profile/
        """

        relative_url = self.app.make("router").route(name, params)
        if absolute:
            return self.url(relative_url)
        else:
            return relative_url



from ..utils.collections import flatten
from ..exceptions import RouteNotFoundException


class Router:
    def __init__(self, *routes, module_location=None):
        self.routes = flatten(routes)

    def find(self, path, request_method, subdomain=None):

        for route in self.routes:
            if route.match(path, request_method, subdomain=subdomain):
                return route

    def matches(self, path):
        for route in self.routes:
            if route.matches(path):
                return route

    def find_by_name(self, name):
        for route in self.routes:
            if route.match_name(name):
                return route

    def route(self, name, parameters={}):
        route = self.find_by_name(name)
        if route:
            return route.to_url(parameters)
        raise RouteNotFoundException(f"Could not find route with the name '{name}'")

    def set_controller_locations(self, location):
        self.controller_locations = location
        return self

    def add(self, *routes):
        self.routes.append(*routes)
        self.routes = flatten(self.routes)

    def set(self, *routes):
        self.routes = []
        self.routes.append(*routes)
        self.routes = flatten(self.routes)

    @classmethod
    def compile_to_url(cls, uncompiled_route, params={}):
        """Compile the route url into a usable url: converts /url/@id into /url/1.
        Used for redirection

        Arguments:
            route {string} -- An uncompiled route like (/dashboard/@user:string/@id:int)
        Keyword Arguments:
            params {dict} -- Dictionary of parameters to pass to the route (default: {{}})
        Returns:
            string -- Returns a compiled string (/dashboard/joseph/1)
        """
        if "http" in uncompiled_route:
            return uncompiled_route

        # Split the url into a list
        split_url = uncompiled_route.split("/")

        # Start beginning of the new compiled url
        compiled_url = "/"

        # Iterate over the list
        for url in split_url:
            if url:
                # if the url contains a parameter variable like @id:int
                if "@" in url:
                    url = url.replace("@", "").split(":")[0]
                    if isinstance(params, dict):
                        compiled_url += str(params[url]) + "/"
                    elif isinstance(params, list):
                        compiled_url += str(params.pop(0)) + "/"
                elif "?" in url:
                    url = url.replace("?", "").split(":")[0]
                    if isinstance(params, dict):
                        compiled_url += str(params.get(url, "/")) + "/"
                    elif isinstance(params, list):
                        compiled_url += str(params.pop(0)) + "/"
                else:
                    compiled_url += url + "/"

        compiled_url = compiled_url.replace("//", "")
        # The loop isn't perfect and may have an unwanted trailing slash
        if compiled_url.endswith("/") and not uncompiled_route.endswith("/"):
            compiled_url = compiled_url[:-1]

        # The loop isn't perfect and may have 2 slashes next to eachother
        if "//" in compiled_url:
            compiled_url = compiled_url.replace("//", "/")

        return compiled_url



class CommandCapsule:
    def __init__(self, command_application):
        self.command_application = command_application
        self.commands = []

    def add(self, *commands):
        self.commands.append(commands)
        self.command_application.add_commands(*commands)
        return self

    def swap(self, command):
        command_name = command.config.name
        # if command with same name has been registered remove it
        if self.command_application.find(command_name):
            # no public API to do this yet
            del self.command_application.commands._commands[command_name]
        self.add(command)

    def run(self):
        return self.command_application.run()



"""Notifications Settings."""
import os

DRIVERS = {
    "slack": {
        "token": os.getenv("SLACK_TOKEN", ""),  # used for API mode
        "webhook": os.getenv("SLACK_WEBHOOK", ""),  # used for webhook mode
        # "mode": os.getenv("SLACK_MODE", "webhook"),  # webhook or api
    },
    "vonage": {
        "key": os.getenv("VONAGE_KEY", ""),
        "secret": os.getenv("VONAGE_SECRET", ""),
        "sms_from": os.getenv("VONAGE_SMS_FROM", "+33000000000"),
    },
    "database": {
        "connection": "sqlite",
        "table": "notifications",
    },
}

DRY = False



"""String generators and helpers"""
import random
import string


def random_string(length=4):
    """Generate a random string based on the given length.

    Keyword Arguments:
        length {int} -- The amount of the characters to generate (default: {4})

    Returns:
        string
    """
    return "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(length)
    )


def modularize(file_path, suffix=".py"):
    """Transforms a file path to a dotted path. On UNIX paths contains / and on Windows \\.

    Keyword Arguments:
        file_path {str} -- A file path such app/controllers

    Returns:
        value {str} -- a dotted path such as app.controllers
    """
    # if the file had the .py extension remove it as it's not needed for a module
    return removesuffix(file_path.replace("/", ".").replace("\\", "."), suffix)


def as_filepath(dotted_path):
    """Inverse of modularize, transforms a dotted path to a file path (with /).

    Keyword Arguments:
        dotted_path {str} -- A dotted path such app.controllers

    Returns:
        value {str} -- a file path such as app/controllers
    """
    return dotted_path.replace(".", "/")


def removeprefix(string, prefix):
    """Implementation of str.removeprefix() function available for Python versions lower than 3.9."""
    if string.startswith(prefix):
        return string[len(prefix) :]
    else:
        return string


def removesuffix(string, suffix):
    """Implementation of str.removesuffix() function available for Python versions lower than 3.9."""
    if suffix and string.endswith(suffix):
        return string[: -len(suffix)]
    else:
        return string



from tests import TestCase
from src.masonite.notification import Notification, Notifiable
from src.masonite.mail import Mailable
from masoniteorm.models import Model


class User(Model, Notifiable):
    """User Model"""

    __fillable__ = ["name", "email", "password"]


class WelcomeUserNotification(Notification):
    def to_mail(self, notifiable):
        return (
            Mailable()
            .to(notifiable.email)
            .subject("Masonite 4")
            .from_("joe@masoniteproject.com")
            .text(f"Hello {notifiable.name}")
        )

    def via(self, notifiable):
        return ["mail"]


class WelcomeNotification(Notification):
    def to_mail(self, notifiable):
        return (
            Mailable()
            .subject("Masonite 4")
            .from_("joe@masoniteproject.com")
            .text("Hello from Masonite!")
        )

    def via(self, notifiable):
        return ["mail"]


class TestMailDriver(TestCase):
    def setUp(self):
        super().setUp()
        self.notification = self.application.make("notification")

    def test_send_to_anonymous(self):
        self.notification.route("mail", "test@mail.com").send(WelcomeNotification())

    def test_send_to_notifiable(self):
        user = User.find(1)
        user.notify(WelcomeUserNotification())

    def test_send_and_override_driver(self):
        # TODO: but I don't really know how to proceed as driver can't be defined anymore
        # in the Mailable
        # Some API solutions:
        # self.notification.route("mail", "test@mail.com").send(WelcomeNotification()).driver("log")
        # self.notification.route("mail", "test@mail.com").send(WelcomeNotification(), driver="log")
        # self.notification.route("mail", "test@mail.com", driver="log").send(WelcomeNotification())
        pass



from tests import TestCase
from src.masonite.routes import Route


class TestHttpRequests(TestCase):
    def test_csrf_request(self):
        self.withoutCsrf()
        return self.post("/").assertContains("Welcome")

    def setUp(self):
        super().setUp()
        self.addRoutes(
            Route.get("/test-with-input", "WelcomeController@form_with_input"),
        )

    def test_old_helper(self):
        self.get("/test-with-input", {"name": "Sam"})
        assert self.application.make("request").old("name") == "Sam"
        assert self.application.make("request").old("wrong-input") == ""



"""The Masonite Response Object."""

import json
import mimetypes
from pathlib import Path

from ..routes.Router import Router
from ..exceptions import ResponseError, InvalidHTTPStatusCode
from ..headers import HeaderBag, Header
from ..utils.http import HTTP_STATUS_CODES
from ..cookies import CookieJar


class Response:
    """A Response object to be used to abstract the logic of getting a response ready to be returned.

    Arguments:
        app {masonite.app.App} -- The Masonite container.
    """

    def __init__(self, app):
        self.app = app
        self.content = ""
        self._status = None
        self.statuses = HTTP_STATUS_CODES
        self.header_bag = HeaderBag()
        self.cookie_jar = CookieJar()
        self.original = None

    def json(self, payload, status=200):
        """Gets the response ready for a JSON response.

        Arguments:
            payload {dict|list} -- Either a dictionary or a list.

        Returns:
            string -- Returns a string representation of the data
        """
        self.content = bytes(json.dumps(payload), "utf-8")
        self.make_headers(content_type="application/json; charset=utf-8")
        self.status(status)

        return self.data()

    def make_headers(self, content_type="text/html; charset=utf-8"):
        """Make the appropriate headers based on changes made in controllers or middleware.

        Keyword Arguments:
            content_type {str} -- The content type to set. (default: {"text/html; charset=utf-8"})
        """
        self.header_bag.add(Header("Content-Length", str(len(self.to_bytes()))))

        # If the user did not change it directly
        self.header_bag.add_if_not_exists(Header("Content-Type", content_type))

    def header(self, name, value=None):
        if value is None and isinstance(name, dict):
            for name, value in name.items():
                self.header_bag.add(Header(name, str(value)))
        elif value is None:
            header = self.header_bag.get(name)
            if isinstance(header, str):
                return header
            return header.value

        return self.header_bag.add(Header(name, value))

    def get_headers(self):
        return self.header_bag.render()

    def cookie(self, name, value=None, **options):
        if value is None:
            cookie = self.cookie_jar.get(name)
            if not cookie:
                return
            return cookie.value

        return self.cookie_jar.add(name, value, **options)

    def delete_cookie(self, name):
        self.cookie_jar.delete(name)
        return self

    def get_response_content(self):
        return self.data()

    def status(self, status):
        """Set the HTTP status code.

        Arguments:
            status {string|integer} -- A string or integer with the standardized status code

        Returns:
            self
        """
        if isinstance(status, str):
            self._status = status
        elif isinstance(status, int):
            try:
                self._status = self.statuses[status]
            except KeyError:
                raise InvalidHTTPStatusCode
        return self

    def is_status(self, code):
        return self._get_status_code_by_value(self.get_status_code()) == code

    def _get_status_code_by_value(self, value):
        for key, status in self.statuses.items():
            if status == value:
                return key

        return None

    def get_status_code(self):
        """Gets the HTTP status string like "200 OK"

        Returns:
            self
        """
        return self._status

    def get_status(self):
        return self._get_status_code_by_value(self.get_status_code())

    def data(self):
        """Get the data that will be returned to the WSGI server.

        Returns:
            string -- Returns a string representation of the response
        """
        if isinstance(self.content, str):
            return bytes(self.content, "utf-8")

        return self.content

    def converted_data(self):
        """Converts the data appropriately so the WSGI server can handle it.

        Returns:
            string -- Returns a string representation of the data
        """
        if isinstance(self.data(), (dict, list)):
            return json.dumps(self.data())
        else:
            return self.data()

    def view(self, view, status=200):
        """Set a string or view to be returned.

        Arguments:
            view {string|dict|list|masonite.view.View} -- Some data type that is an appropriate response.

        Keyword Arguments:
            status {int} -- The Response status code. (default: {200})

        Raises:
            ResponseError -- If a data type that is not an acceptable response type is returned.

        Returns:
            string|dict|list -- Returns the data to be returned.
        """
        self.original = view

        if isinstance(view, tuple):
            view, status = view
            self.status(status)

        if not self.get_status_code():
            self.status(status)

        if isinstance(view, (dict, list)):
            return self.json(view, status=self.get_status_code())
        elif hasattr(view, "serialize"):
            return self.json(view.serialize(), status=self.get_status_code())
        elif isinstance(view, int):
            view = str(view)
        elif hasattr(view, "get_response"):
            view = view.get_response()
        elif view is None:
            raise ResponseError(
                "Responses cannot be of type: None. Did you return anything in your responsable method?"
            )

        self.content = view

        self.make_headers()

        return self

    def back(self):
        return self.redirect(url=self.app.make("request").get_back_path())

    def redirect(self, location=None, name=None, params={}, url=None, status=302):
        """Set the redirection on the server.

        Keyword Arguments:
            location {string} -- The URL to redirect to (default: {None})
            status {int} -- The Response status code. (default: {302})
            params {dict} -- The route params (default: {})

        Returns:
            string -- Returns the data to be returned.
        """
        self.status(status)

        if location:
            self.header_bag.add(Header("Location", location))
        elif name:
            url = self._get_url_from_route_name(name, params)
            self.header_bag.add(Header("Location", url))
        elif url:
            self.header_bag.add(Header("Location", url))
        self.view("Redirecting ...")
        return self

    def _get_url_from_route_name(self, name, params={}):
        route = self.app.make("router").find_by_name(name)
        if not route:
            raise ValueError(f"Route with the name '{name}' not found.")
        return Router.compile_to_url(route.url, params)

    def to_bytes(self):
        """Converts the data to bytes so the WSGI server can handle it.

        Returns:
            bytes -- The converted response to bytes.
        """
        return self.converted_data()

    def download(self, name, location, force=False):
        if force:
            self.header("Content-Type", "application/octet-stream")
            self.header(
                "Content-Disposition",
                'attachment; filename="{}{}"'.format(name, self.extension(location)),
            )
        else:
            self.header("Content-Type", self.mimetype(location))

        with open(location, "rb") as filelike:
            data = filelike.read()

        return self.view(data)

    def extension(self, path):
        return Path(path).suffix

    def mimetype(self, path):
        """Gets the mimetime of a path

        Arguments:
            path {string} -- The path of the file to download.

        Returns:
            string -- The mimetype for use in headers
        """
        return mimetypes.guess_type(path)[0]
